import re
import inflection
import functools
from . import MungeError

TREATMENT_QUALIFIER_TYPES = ['Amount', 'Temperature', 'Duration']


def get_defaults_callback(template):
    """
    Returns the callback function to set default output row values.

    The callback function takes two parameters:
    * `row` - the output row
    * `out_col_ndx_map` - the output column {name: index} dictionary

    The callback function modifies the given row in place. It does
    not return a value.

    *Note*: although the callback raises an exception if certain
    values necessary to make a default are missing, it does not
    fully validate the output row. The :mod:`validators` module
    serves that purpose.

    :param template: the template name
    :return: the callback function to set default values, or `None`
        if there is no defaults function for the given template
    :raise MungeError: if an output row column value required to
        add a default is missing
    """
    callback = DEF_CALLBACKS.get(template)
    if callback:
        return callback
    elif '.' in template:
        # The experiment type, e.g. ELISA.
        prefix, exp_type = template.split('.')
        callback = DEF_CALLBACKS.get(prefix)
        if callback:
            return functools.partial(callback, experiment_type=exp_type)


def _add_protocols_defaults(row, out_col_ndx_map, experiment_type=None):
    """
    Makes default values as necessary for the following required columns:
    * `Name` - `File Name` without extension formatted as a title
    * `Description` - copied from `Name`
    * `User Defined ID` - lower-case, underscored `File Name`
    """
    file_name_ndx = out_col_ndx_map['File Name']
    file_name = row[file_name_ndx]
    if not file_name:
        raise MungeError("The File Name value is missing")
    file_name_sans_ext = re.sub('\..*$', '', file_name)
    name_ndx = out_col_ndx_map['Name']
    name = row[name_ndx]
    if not name:
        name = row[name_ndx] = _titleize(file_name_sans_ext)
    desc_ndx = out_col_ndx_map['Description']
    if not row[desc_ndx]:
        row[desc_ndx] = name
    user_id_ndx = out_col_ndx_map['User Defined ID']
    if not row[user_id_ndx]:
        row[user_id_ndx] = (
            inflection.underscore(file_name_sans_ext).replace(' ', '_')
        )


def _add_reagents_defaults(row, out_col_ndx_map, experiment_type=None):
    """
    Makes default values as necessary for the following required columns:
    * `Name` - derived from the `Analyte Reported` and `Catalog Number`
    * `Catalog Number` - `NA`
    * `User Defined ID` - lower-case `Catalog Number` followed by `_reagent`
    * `Description` copied from `Name`
    """
    cat_nbr_ndx = out_col_ndx_map['Catalog Number']
    cat_nbr = row[cat_nbr_ndx]
    if not cat_nbr:
        cat_nbr = row[cat_nbr_ndx] = 'NA'
    analyte = row[out_col_ndx_map['Analyte Reported']]
    if not analyte:
        raise MungeError("Required Analyte Reported value is missing")
    # The default name is used for both the name and the default
    # user id.
    if cat_nbr.lower() == 'na':
        name_parts = [analyte]
        if experiment_type:
            name_parts.append(experiment_type)
    else:
        name_parts = [cat_nbr]
    name_parts.append('Reagent')
    def_name = ' '.join(name_parts)
    name_ndx = out_col_ndx_map['Name']
    name = row[name_ndx]
    if not name:
        name = row[name_ndx] = def_name
    # The default user id is derived from the name.
    user_id_ndx = out_col_ndx_map['User Defined ID']
    if not row[user_id_ndx]:
        row[user_id_ndx] = _name_to_id(def_name)
    # The default description is the name.
    desc_ndx = out_col_ndx_map['Description']
    if not row[desc_ndx]:
        row[desc_ndx] = name


def _add_treatments_defaults(row, out_col_ndx_map, experiment_type=None):
    """
    Makes default values as necessary for the following required columns:
    * `Name` - derived from the treatment values
    * `User Defined ID` - lower-case, underscored `Name` and `Amount Value`
    * `Use Treatment?` - default is `Yes`
    """
    name_ndx = out_col_ndx_map['Name']
    name = row[name_ndx]
    if not name:
        name = _default_treatment_name(row, out_col_ndx_map)
        if not name:
            raise MungeError("Required Name value could not be inferred from " +
                             "the " + ','.join(attrs) + " attributes")
        row[name_ndx] = name
    user_id_ndx = out_col_ndx_map['User Defined ID']
    if not row[user_id_ndx]:
        row[user_id_ndx] = _name_to_id(name)
    use_treatment_ndx = out_col_ndx_map['Use Treatment?']
    if not row[use_treatment_ndx]:
        row[use_treatment_ndx] = 'Yes'


def _default_treatment_name(row, out_col_ndx_map, experiment_type=None):
    qualifiers = [_treatment_name_qualifier(row, out_col_ndx_map, type)
                  for type in TREATMENT_QUALIFIER_TYPES]
    qualifier_s = ', '.join(filter(None, qualifiers))
    return qualifier_s if qualifier_s else None


def _treatment_name_qualifier(row, out_col_ndx_map, treatment_type):
    value_ndx = out_col_ndx_map[treatment_type + ' Value']
    value = row[value_ndx]
    if value != None:
        unit_ndx = out_col_ndx_map[treatment_type + ' Unit']
        unit = row[unit_ndx]
        if unit != 'Not Specified':
            return ' '.join((value, unit))


def _add_subjectAnimals_defaults(row, out_col_ndx_map):
    """
    Makes default values as necessary for the following required columns:
    * `Age Unit` - `Days`
    * `Age Event` - `Age at infection`
    """
    age_unit_ndx = out_col_ndx_map['Age Unit']
    age_unit = row[age_unit_ndx]
    if not age_unit:
        row[age_unit_ndx] = 'Days'
    age_event_ndx = out_col_ndx_map['Age Event']
    age_event = row[age_event_ndx]
    if not age_event:
        row[age_event_ndx] = 'Age at infection'


def _add_samples_defaults(row, out_col_ndx_map, experiment_type=None):
    """
    Makes default values as necessary for the following required columns:
    * `Experiment ID` - lower-case, underscored `Experiment Name`
    * `Planned Visit ID` - `Study ID` followed by `d` and the
      `Study Time Collected`
    * `Biosample ID` - lower-case, underscored `Biosample Name`,
      if present, otherwise the `Expsample ID`, if present,
      otherwise derived from the `Subject ID`, `Treatment ID(s)`
      and `Experiment ID`
    * `Expsample ID` - derived from the `Biosample ID`, `Treatment ID`
      and Experiment ID
    """
    experiment_id_ndx = out_col_ndx_map['Experiment ID']
    if not row[experiment_id_ndx]:
        name_ndx = out_col_ndx_map['Experiment Name']
        name = row[name_ndx]
        if not name:
            msg = "Both the Experiment ID and Name values are missing"
            raise MungeError(msg)
        row[experiment_id_ndx] = _name_to_id(name)
    visit_id_ndx = out_col_ndx_map['Planned Visit ID']
    visit_id = row[visit_id_ndx]
    if not visit_id:
        visit_id = _default_visit_id(out_col_ndx_map, 'Study Time Collected',
                                     row)
        row[visit_id_ndx] = visit_id

    biosample_id_ndx = out_col_ndx_map['Biosample ID']
    biosample_id = row[biosample_id_ndx]
    if not biosample_id:
        biosample_id = _default_biosample_id(row, out_col_ndx_map)
        row[biosample_id_ndx] = biosample_id
    expsample_id_ndx = out_col_ndx_map['Expsample ID']
    if not row[expsample_id_ndx]:
        row[expsample_id_ndx] = biosample_id


def _default_biosample_id(row, out_col_ndx_map):
    expsample_id_ndx = out_col_ndx_map['Expsample ID']
    expsample_id = row[expsample_id_ndx]
    if expsample_id:
        return expsample_id
    name_ndx = out_col_ndx_map['Biosample Name']
    name = row[name_ndx]
    if name:
        return _name_to_id(name)
    subject_id_ndx = out_col_ndx_map['Subject ID']
    subject_id = row[subject_id_ndx]
    if not subject_id:
        raise MungeError("The Biosample ID, Expsample ID, Biosample Name" +
                         " and Subject ID values are missing")
    experiment_id_ndx = out_col_ndx_map['Experiment ID']
    experiment_id = row[experiment_id_ndx]
    if not experiment_id:
        raise MungeError("The Biosample ID, Expsample ID, Biosample Name" +
                         " and Experiment ID values are missing")
    time_ndx = out_col_ndx_map['Study Time Collected']
    time = row[time_ndx]
    if not time:
        raise MungeError("The Biosample ID, Expsample ID, Biosample Name" +
                         " and Study Time Collected values are missing")
    time_unit_ndx = out_col_ndx_map['Study Time Collected Unit']
    time_unit = row[time_unit_ndx]
    if not time_unit:
        raise MungeError("The Biosample ID, Expsample ID, Biosample Name" +
                         " and Study Time Collected Unit values are missing")
    treatment_ids_ndx = out_col_ndx_map['Treatment ID(s)']
    treatment_ids = row[treatment_ids_ndx]
    if not treatment_ids:
        raise MungeError("The Biosample ID, Expsample ID, Biosample Name" +
                         " and Treatment ID(s) values are missing")

    return '_'.join((subject_id, re.sub(', *', '_', treatment_ids),
                     time_unit[0].lower() + time, experiment_id))


def _default_visit_id(out_col_ndx_map, study_day_column, row, experiment_type=None):
    """
    :return: the `Study ID` followed by `d` and the study day
    """
    study_id = row[out_col_ndx_map['Study ID']]
    if not study_id:
        raise MungeError("Required Study ID value is missing")
    day = row[out_col_ndx_map[study_day_column]]
    if day == None:
        raise MungeError("Required %s value is missing" % study_day_column)
    day_id = 'd' + str(day)
    return '_'.join((study_id.lower(), day_id))


def _add_assessments_defaults(row, out_col_ndx_map, experiment_type=None):
    """
    Makes default values as necessary for the following required columns:
    * `Planned Visit ID` - `Study ID` followed by `d` and the `Study Day`
    * `Panel Name Reported` - copied from the `Assessment Type`
    * `Assessment Panel ID` - derived from the `Panel Name Reported`
    * `User Defined ID` - derived from the `Subject ID`, `Planned Visit ID`
      and `Component Name Reported`
    """
    visit_id_ndx = out_col_ndx_map['Planned Visit ID']
    visit_id = row[visit_id_ndx]
    if not visit_id:
        visit_id = _default_visit_id(out_col_ndx_map, 'Study Day', row)
        row[visit_id_ndx] = visit_id
    panel_name_ndx = out_col_ndx_map['Panel Name Reported']
    panel_name = row[panel_name_ndx]
    if not panel_name:
        type = row[out_col_ndx_map['Assessment Type']]
        if not type:
            raise MungeError("Neither the Panel Name Reported nor the " +
                             "Assessment Type value is missing")
        panel_name = row[panel_name_ndx] = type
    panel_id_ndx = out_col_ndx_map['Assessment Panel ID']
    panel_id = row[panel_id_ndx]
    if not panel_id:
        panel_id = row[panel_id_ndx] = panel_name.lower().replace(' ', '_')
    user_id_ndx = out_col_ndx_map['User Defined ID']
    if not row[user_id_ndx]:
        subject_id = row[out_col_ndx_map['Subject ID']]
        if not subject_id:
            raise MungeError("Required Subject ID value is missing")
        component_name_ndx = out_col_ndx_map['Component Name Reported']
        component_name = row[component_name_ndx]
        if not component_name:
            raise MungeError("Required Component Name Reported value is missing")
        row[user_id_ndx] = '_'.join((subject_id, visit_id, component_name.lower()))


def _name_to_id(name):
    return re.sub(r'[^\w]', '_', name).lower()


def _titleize(s):
    """
    Tweaks `inflection.titleize` to preserve acronyms.

    >>> print(inflection.titleize('AB_CdeXYZ_Name'))
    Ab Cde Xyz Name

    >>> print(defaults._titleize('AB_CdeXYZ_Name'))
    AB Cde XYZ Name
    """
    acronyms = re.findall('[A-Z][A-Z]+', s)
    recapitalize = {acronym.capitalize(): acronym for acronym in acronyms}
    title = inflection.titleize(s)
    for k, v in recapitalize.items():
        title = title.replace(k, v)
    return title


"""The default callbuck functions."""
DEF_CALLBACKS = dict(protocols=_add_protocols_defaults,
                     treatments=_add_treatments_defaults,
                     reagents=_add_reagents_defaults,
                     subjectAnimals=_add_subjectAnimals_defaults,
                     experimentSamples=_add_samples_defaults,
                     assessments=_add_assessments_defaults)



import functools
from . import MungeError


class ValidationError(MungeError):
    pass


def get_validator(template):
    """
    Returns the callback function to validate output rows.

    The callback function takes two parameters:
    * `out_col_ndx_map` - the output column {name: index} dictionary
    * `row` - the output row

    The callback function modifies the given row in place. It does
    not return a value.

    *Note*: although the callback raises an exception if certain
    values necessary to make a default are missing, it does not
    fully validate the output row. The :mod:`validators` module
    serves that purpose.
    :param template: the template name
    :return: the validator, if any
    """
    required = _get_required(template)
    if required:
        return functools.partial(_validate, required)


def _get_required(template):
    """
    :param template: the template name
    :return: the required columns, if any
    """
    required = REQUIRED_COLUMNS.get(template)
    if required:
        return required
    elif '.' in template:
        prefix = template.split('.')[0]
        return REQUIRED_COLUMNS.get(prefix)


def _validate(required, row, out_col_ndx_map):
    missing = [col for col in required
               if row[out_col_ndx_map[col]] == None]
    if missing:
        raise ValidationError("Missing values for columns: %s" % missing)


"""The validation functions."""
REQUIRED_COLUMNS = dict(
    subjectAnimals=['Subject ID', 'Arm Or Cohort ID', 'Gender',
                    'Min Subject Age', 'Age Unit', 'Age Event',
                    'Subject Location', 'Species', 'Strain'],
    subjectHumans=['Subject ID', 'Arm Or Cohort ID', 'Gender',
                   'Min Subject Age', 'Age Unit', 'Age Event',
                   'Subject Location', 'Ethnicity', 'Race'],
    protocols=['User Defined ID', 'File Name', 'Name'],
    experimentSamples=['Study ID', 'Expsample ID', 'Reagent ID(s)',
                       'Treatment ID(s)', 'Biosample ID', 'Subject ID',
                       'Planned Visit ID', 'Study Time Collected',
                       'Study Time Collected Unit', 'Study Time T0 Event',
                       'Protocol ID(s)', 'Experiment ID', 'Experiment Name',
                       'Measurement Technique'],
    treatments=['User Defined ID', 'Name', 'Use Treatment?'],
    reagents=['User Defined ID', 'Manufacturer', 'Catalog Number',
              'Analyte Reported'],
    assessments=['Subject ID', 'Assessment Panel ID', 'Study ID',
                 'Panel Name Reported', 'Component Name Reported',
                 'User Defined ID', 'Planned Visit ID', 'Study Day'],
    FCM_Derived_data=['Expsample ID', 'Population Name Reported',
                      'Gating Definition Reported',
                      'Population Statistic (count, percentile, etc)',
                      'Population Stat Unit Reported', 'Workspace File'])

# ELISA samples require values.
REQUIRED_COLUMNS['experimentSamples.ELISA'] = (
    REQUIRED_COLUMNS['experimentSamples'] +
    ['Analyte Reported', 'Value Reported', 'Unit Reported'])

# Flow samples require the .fcs file name.
REQUIRED_COLUMNS['experimentSamples.Flow_Cytometry'] = (
    REQUIRED_COLUMNS['experimentSamples'] + ['.Fcs Result File'])




from bunch import Bunch
import requests
from . import MungeError

# The templates location.
URL_DIR = 'https://www.immport.org/downloads/data/upload/templates/txt-templates'

# The template name: mapping dictionary.
tmpl_dict = {}


class TemplateError(MungeError):
    pass


def assessments():
    """
    Fetches the assessments template on demand.

    :return: the assessments template {header, columns} dictionary
    """
    return template_for('assessments')


def template_for(name):
    """
    Fetches the given template on demand.

    :param name: the template name
    :return: the template {header, columns} dictionary
    """
    mapping = tmpl_dict.get(name)
    if not mapping:
        mapping = tmpl_dict[name] = _load(name)

    return mapping


def _load(name):
    """
    Loads the given template from the Immport web site.

    :param name: the template name
    :return: the template {header, columns} dictionary
    """
    url = URL_DIR + '/' + name + '.txt'
    try:
        response = requests.get(url, stream=True)
        response.raise_for_status()
    except Exception as e:
        msg = "The online template could not be accessed: " + url
        raise TemplateError(msg) from e
    rows = [line.decode('utf-8').split("\t") for line in response.iter_lines()]
    cols_row_ndx = next((i for i, row in enumerate(rows)
                         if row[0] == 'Column Name'), -1)
    if cols_row_ndx == -1:
        raise TemplateError("The online template %s could not be parsed: %s" %
                            (name, url))
    header = rows[0:cols_row_ndx]
    columns = rows[cols_row_ndx]
    # Special case: assessments has a duplicate column name for
    # different purposes.
    if name == 'assessments':
        indexes = [index for index, col in enumerate(columns)
                   if col == 'Name Reported']
        if len(indexes) == 2:
            columns[indexes[0]] = 'Panel Name Reported'
            columns[indexes[1]] = 'Component Name Reported'
        aliases = {'Panel Name Reported': 'Name Reported',
                   'Component Name Reported': 'Name Reported'}
    else:
        aliases = {}

    return Bunch(name=name, header=header, columns=columns, aliases=aliases)



import os
import re
import glob
import warnings
from setuptools import (setup, find_packages)


class InstallError(Exception):
    """imppload installation error."""
    pass


def version(package):
    """
    :return: the package version as listed in the package `__init.py__`
        `__version__` variable.
    """
    # The version string parser.
    REGEXP = re.compile("""
       __version__   # The version variable
       \s*=\s*       # Assignment
       ['\"]         # Leading quote
       (.+)          # The version string capture group
       ['\"]         # Trailing quote
    """, re.VERBOSE)

    with open(os.path.join(package, '__init__.py')) as fd:
       match = REGEXP.search(fd.read())
       if not match:
           raise InstallError("The %s __version__ variable was not found" %
                              package)
       return match.group(1)


def readme():
    with open("README.md") as fd:
        return fd.read()


setup(
    name = 'immpload',
    version = version('immpload'),
    author = 'Oregon Health & Sciences University',
    author_email = 'loneyf@ohsu.edu',
    platforms = 'Any',
    license = 'MIT',
    keywords = 'Immpload',
    packages = find_packages(exclude=['test**']),
    package_data = dict(immpload=['conf/*.yaml']),
    scripts = glob.glob('bin/*'),
    url = 'https://github.com/biodev/imppload/',
    description = 'Immport upload preparation',
    long_description_content_type='text/markdown',
    long_description = readme(),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    install_requires = ['bunch', 'PyYAML', 'openpyxl', 'inflection', 'deepmerge']
)



import os
import csv
import re
import yaml
from bunch import Bunch
import openpyxl
import functools
from deepmerge import always_merger
from . import (MungeError, templates, defaults, validators)

"""The pattern to detect a config configuration pattern."""
PATTERN_PAT = re.compile("r['\"](?P<pattern>.+)['\"]$")


def munge(template, *in_files, config=None, out_dir=None, sheet=None,
          input_filter=None, callback=None, defaults_opt=True,
          validate_opt=True, append_opt=False, **kwargs):
    """
    Builds the Immport upload file for the given input file.
    The template is a supported Immport template name, e.g.
    `assessments`. The output is the Immport upload file,
    e.g. `assessments,txt`, placed in the output directory.

    The keyword arguments (_kwargs_) are static output
    _column_`=`_value_ definitions that are applied to every
    output row. The column name can be underscored, e.g.
    `Study_ID`.

    Output validation is disabled by default, but recommended
    for new uploads. Enable validation by setting the _validate_
    flag parameter to `True`.

    :param template: the required Immport template name
    :param in_files: the input file(s) to munge
    :param config: the configuration dictionary or file name
        of list of file names
    :param out_dir: the target location (default current directory)
    :param sheet: for an Excel workbook input file, the sheet to open
    :param input_filter: optional input row validator which has
        parameter in_row and returns whether the row is valid
    :param callback: optional callback with parameters
        in_row, in_col_ndx_map, out_col_ndx_map and out_row returning
        an array of rows to write to the output file
    :param defaults_opt: flag indicating whether to add defaults to the
        output (default `True`)
    :param validate_opt: flag indicating whether to validate the
        output for required fields (default `True`)
    :param append_opt: append rather than overwrite an existing output
        file (default False)
    :param kwargs: the optional static _column_`=`_value_ definitions
    :return: the output file name
    """
    # Note: this method signature and apidoc is copied in the
    # package README.md file.

    # Load the config file, if necessary.
    if config == None:
        config = {}
    else:
        if isinstance(config, str):
            config_files = [config]
        elif isinstance(config, list):
            config_files = config
        else:
            msg = ("Unsupported config parameter - expected" +
                   " dictionary, string or list of strings: %s" % config)
            raise MungeError(msg)
        config = {}
        for f in config_files:
            with open(f, 'r') as fd:
                always_merger.merge(config, yaml.safe_load(fd))
    # Parse the config to a {name, value} bunch.
    config = _parse_config(config)
    # Collect options into the config.
    config.in_filter = input_filter
    # Capture the append flag.
    config.append = append_opt

    # Remove the template file suffix, if necessary.
    template = re.sub("\.txt$", '', template)
    # Acquire the template object.
    config.template = templates.template_for(template)
    # Make the output column {name: index} dictionary.
    config.out_col_ndx_map = {col: i
                              for i, col in enumerate(config.template.columns)}

    # Add the kwargs to the static config.
    static_kwargs = {key.replace('_', ' '): value
                     for key, value in kwargs.items()}
    config.static.update(static_kwargs)
    # Validate the static definition.
    for out_col in config.static:
        _validate_static_column(out_col, config.template)

    # Add the default template callback, if necessary.
    if defaults_opt:
        def_callback = defaults.get_defaults_callback(template)
        if def_callback:
            callback = functools.partial(_add_defaults, def_callback,
                                         callback=callback)
    # Add validation, if necessary.
    if validate_opt:
        validator = validators.get_validator(template)
        if validator:
            callback = functools.partial(_validate_output, validator,
                                         callback=callback)
    # Set the composite configuration callback.
    config.callback = callback

    # Prepare the output location.
    if out_dir:
        out_dir = os.path.abspath(out_dir)
        os.makedirs(out_dir, exist_ok=True)
    else:
        out_dir = os.getcwd()
    out_file = os.path.join(out_dir, template + '.txt')

    # Convert the input to output.
    mode = 'a' if append_opt else 'w'
    with open(out_file, mode) as fd:
        writer = csv.writer(fd, delimiter='\t')
        if in_files:
            for i, in_file in enumerate(in_files):
                reader = _create_reader(in_file, sheet=sheet)
                _munge_file(in_file, reader, writer, config)
                # Subsequent files append the content.
                if i == 0:
                    config.append = True
        else:
            # Only use the configuration to generate output.
            _munge_config(writer, config)

    return out_file


def _validate_static_column(out_col, template):
    if not out_col in template.columns:
        # Check for a column name alias.
        aliases = [col for col, alias in template.aliases.items()
                   if alias == out_col]
        if aliases:
            msg = (("Output column '%s' must be disambiguated as " +
                    "one of the following: %s") % (out_col, aliases))
        else:
            msg = "Output column not found: %s" % out_col
        raise MungeError(msg)


def _is_valid_input_row(row, conf_filter=None):
    """
    Filters blank rows and applies the optional configuration filter
    argument if necessary.

    :param row: the input row
    :param conf_filter: the optional template-specific configuration filter
    :return: whether the row should be converted
    """
    # Filter blank rows.
    if all(value == None for value in row):
        return False
    # Apply the config filter, if any.
    return conf_filter(row) if conf_filter else True


def _validate_output(validator, in_row, in_col_ndx_map, out_col_ndx_map,
                     out_row, callback=None):
    if callback:
        out_rows = callback(in_row, in_col_ndx_map, out_col_ndx_map, out_row)
    else:
        out_rows = [out_row]
    for row in out_rows:
        validator(row, out_col_ndx_map)

    return out_rows


def _parse_config(config):
    """
    Converts the input config to a
    {names, patterns, static, keys, values_dict} bunch, where
    _names_ is the {input column: output column} dictionary,
    _patterns_ is the {pattern: output column} dictionary,
    _static_ is the {output column: value} dictionary,
    _key_ is the [output columns] uniqueness list, and
    *value_dict* is the {output column: value conversion} dictionary.
    """
    # Convert the configuration {output: spec} dictionary to
    # {input: name} and {input: value} dictionaries.
    out_in_columns = config.get('columns', {})
    # Partition the config columns and patterns.
    out_in_pat_dict = config.get('patterns', {})
    out_in_patterns = {out_col: _compile_config_pattern(pattern, out_col)
                       for out_col, pattern in out_in_pat_dict.items()}
    conf_values = config.get('values', {})
    values_dict = {out_col: value for out_col, value in conf_values.items()
                   if isinstance(value, dict)}
    static = {out_col: value for out_col, value in conf_values.items()
              if out_col not in values_dict}

    # Invert the {out: in} dictionaries.
    in_out_names = {in_col: [] for in_col in out_in_columns.values()}
    for out_col, in_col in out_in_columns.items():
        in_out_names[in_col].append(out_col)
    in_out_patterns = {pattern: out_col
                       for out_col, pattern in out_in_patterns.items()}

    return Bunch(names=in_out_names, patterns=in_out_patterns,
                 static=static, values_dict=values_dict)


def _compile_config_pattern(pattern, out_col):
    try:
        return re.compile(pattern)
    except re.error as e:
        msg = "Configuration %s pattern %s error" % (out_col, pattern)
        raise MungeError() from e


def _munge_file(in_file, reader, writer, config):
    in_cols = None
    for i, in_row in enumerate(reader):
        # The first (possibly filtered) input file row is
        # the input column names.
        if i == 0:
            # Capture the input column names and indexes.
            _capture_input_columns(in_row, config, writer)
            # If the output is entirely determined by column name matching,
            # then generate the output and skip the remaining lines.
            if _is_column_match_only(config):
                _munge_row(in_file, in_row, i, config, writer)
                return
        elif _is_valid_input_row(in_row, config.in_filter):
            _munge_row(in_file, in_row, i, config, writer)


def _is_column_match_only(config):
    """
    Returns whether the output generation is determined solely by
    extracting values from the input column name first row.

    :return: whether the configuration *names* and *patterns*
        dictionaries are non-empty and map to the same values
    """
    if not config.names:
        return False
    # Flatten the output column lists into a set.
    out_cols = {col for cols in config.names.values() for col in cols}
    return out_cols == set(config.patterns.values())

def _capture_input_columns(in_row, config, writer):
    # Capture the input column names and indexes.
    config.in_col_ndx_map = {name: i for i, name in enumerate(in_row)}
    # Auto-convert the input columns which match
    # output column names.
    for in_col in config.in_col_ndx_map:
        if (in_col in config.out_col_ndx_map and
            not in_col in config.names):
            config.names[in_col] = [in_col]
    # Write the Immport template header lines.
    if not config.append:
        _write_header(config.template, writer)
    # Parse the config patterns, now that we have input
    # columns to match against.
    _make_embedded_config(config)


def _munge_row(in_file, in_row, index, config, writer):
    # The output row generator.
    out_rows = _generate_output(in_row, config)
    # Write the generated rows.
    try:
        writer.writerows(out_rows)
    except Exception as e:
        msg = ("Error generating %s output for input row %d in %s" %
               (config.template.name, index + 1, in_file))
        raise MungeError(msg) from e

def _munge_config(writer, config):
    """
    Generates output based solely on the configuration.
    """
    # Make an empty input {column: index} and embedded map.
    config.in_col_ndx_map = {}
    config.embedded = {}
    # Write the Immport template header lines.
    if not config.append:
        _write_header(config.template, writer)
    # The output row generator.
    out_rows = _generate_output([], config)
    # Write the generated rows.
    try:
        writer.writerows(out_rows)
    except Exception as e:
        msg = "Error generating %s output" % config.template.name
        raise MungeError(msg) from e


def _write_header(template, writer):
    # Write the Immport template header lines.
    _write_prequel(template, writer)
    # Write the template column names.
    _write_column_names(template, writer)


def _make_embedded_config(config):
    """
    Matches config patterns against the input columns to create a
    new config attribute _embedded_ holding a list of
    (output column index: matches) tuples, where _matches_ is a list
    of (input column index, match dictionary) tuples.
    """
    # Make the {output column: {output column: value}} dictionary.
    config.embedded = {out_col: _match_input_columns(pattern, config)
                       for pattern, out_col in config.patterns.items()}


def _match_input_columns(pattern, config):
    """
    :return: the {input column: {output column: value}} dictionary
    """
    matches = [(in_col, pattern.match(in_col))
               for in_col in config.in_col_ndx_map]
    return {in_col: _parse_group_dictionary(match.groupdict(), config)
            for in_col, match in matches if match}


def _parse_group_dictionary(match_dict, config):
    """
    :return: the {output column: value} dictionary
    """
    cols_values = [(config.names[name], value)
                   for name, value in match_dict.items()
                   if name in config.names]
    return {col: value
            for cols, value in cols_values
            for col in cols}


def _create_reader(in_file, sheet=None):
    if not os.path.exists(in_file):
        raise IOError("Input file not found: %s" % in_file)
    _, ext = os.path.splitext(in_file)
    if ext == '.csv' or ext == '.txt':
        reader = _read_csv(in_file)
    elif ext == '.xlsx':
        reader = _read_excel(in_file, sheet=sheet)
    else:
        raise ValueError("Input file type not supported: %s" % in_file)

    return reader


def _write_prequel(template, writer):
    # Write the Immport template header lines.
    for hdg in template.header:
        writer.writerow(hdg)


def _write_column_names(template, writer):
    # Write the template column names.
    columns = template.columns.copy()
    for col, alias in template.aliases.items():
        index = columns.index(col)
        columns[index] = alias
    writer.writerow(columns)


def _generate_output(in_row, config):
    # Map the columns.
    out_row = _map_row(in_row, config)
    # Expand each embedded config into a separate row.
    if config.embedded:
        out_rows = _generate_embedded_output(in_row, out_row, config)
    else:
        out_rows = (out_row,)
    for out_row in out_rows:
        if config.callback:
            for row in _apply_callback(in_row, config, out_row):
                yield row
        else:
            yield out_row


def _apply_callback(in_row, config, out_row):
    return config.callback(in_row, config.in_col_ndx_map,
                           config.out_col_ndx_map, out_row)


def _map_row(in_row, config):
    # Map the output columns to output values.
    cols_values_tuples = [(config.names[col], _map_values(col, in_row[i], config))
                          for i, col in enumerate(config.in_col_ndx_map)
                          if col in config.names]
    col_value_dict = {col: values[i]
                      for cols, values in cols_values_tuples
                      for i, col in enumerate(cols)}
    out_row = [col_value_dict.get(col) for col in config.out_col_ndx_map]
    # Set the statically defined column values.
    for out_col, value in config.static.items():
        # Get the static definition output column index.
        out_col_ndx = config.out_col_ndx_map[out_col]
        out_row[out_col_ndx] = value

    return out_row


def _map_values(in_col, value, config):
    out_cols = config.names[in_col]
    col_values = [config.values_dict.get(out_col)
                  for out_col in out_cols]
    return [value_dict.get(value, value) if value_dict else value
            for value_dict in col_values]


def _generate_embedded_output(in_row, out_row, config):
    """
    :yield: each output row
    """
    for out_col, matches in config.embedded.items():
        for in_col, group_values in matches.items():
            in_col_ndx = config.in_col_ndx_map[in_col]
            # There must be an input value.
            value = in_row[in_col_ndx]
            if value:
                out_col_ndx = config.out_col_ndx_map[out_col]
                yield _create_embedded_row(value, out_row, out_col_ndx,
                                           config, group_values)


def _create_embedded_row(value, out_row, out_col_ndx, config, group_values):
    """
    Copies the given output row and assigns the given pattern match
    group values to the corresponding output row columns defined by
    the configuration.

    :return: the new output row
    """
    # Make a new output row.
    row = out_row.copy()
    # Set the embedded column.
    row[out_col_ndx] = value
    # Assign the match group values to the associated output column.
    for grp_col, grp_value in group_values.items():
        # Map the output columns to output values.
        value_dict = config.values_dict.get(grp_col)
        value = value_dict.get(grp_value, grp_value) if value_dict else grp_value
        col_ndx = config.out_col_ndx_map[grp_col]
        row[col_ndx] = value

    return row


def _add_defaults(def_callback, in_row, in_col_ndx_map, out_col_ndx_map,
                  out_row, callback=None):
    """
    Assigns defaults to the generated output rows.

    :return: the generated output rows
    :rtype: list
    """
    if callback:
        out_rows = callback(in_row, in_col_ndx_map, out_col_ndx_map, out_row)
    else:
        out_rows = [out_row]
    for row in out_rows:
        def_callback(row, out_col_ndx_map)

    return out_rows


def _read_csv(in_file):
    """
    Reads the given comma- or tab-separated file.

    :param in_file: the `.csv` CSV file or `.txt` TSV file
    :yield: each non-empty input row
    """
    kwargs = {}
    if in_file.endswith('.txt'):
        kwargs['delimiter'] = '\t'
    with open(in_file) as fd:
        reader = csv.reader(fd, **kwargs)
        for row in reader:
            # Skip empty rows.
            if row and any(row):
                yield row


def _read_excel(in_file, sheet=None):
    """
    Reads the given Excel file.

    :param in_file: the `.xslx` workbook file
    :param sheet: the optional worksheet (default is the worksheet
        displayed when the workbook was last saved)
    :yield: each worksheet row as a list of cell values
    """
    wb = openpyxl.load_workbook(in_file, read_only=True)
    ws = wb.get_sheet_by_name(sheet) if sheet else wb.active
    for row in ws:
        if row:
            yield [cell.value for cell in row]
